@@ -2,7 +2,7 @@
  * BIOS Decode
  *
  *   (C) 2000-2002 Alan Cox <alan@redhat.com>
- *   (C) 2002-2003 Jean Delvare <khali@linux-fr>
+ *   (C) 2002-2007 Jean Delvare <khali@linux-fr.org>
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -20,14 +20,14 @@
  *
  *   For the avoidance of doubt the "preferred form" of this code is one which
  *   is in an open unpatent encumbered format. Where cryptographic key signing
- *   forms part of the process of creating an executable the information 
+ *   forms part of the process of creating an executable the information
  *   including keys needed to generate an equivalently functional executable
  *   are deemed to be part of the source code.
  *
  * References:
  *  - DMTF "System Management BIOS Reference Specification"
- *    Version 2.3.3
- *    http://www.dmtf.org/standards/bios.php.
+ *    Version 2.3.4
+ *    http://www.dmtf.org/standards/smbios
  *	- Intel "Preboot Execution Environment (PXE) Specification"
  *    Version 2.1
  *    http://www.intel.com/labs/manage/wfm/wfmspecs.htm
@@ -46,38 +46,39 @@
  *  - Compaq "Technical Reference Guide for Compaq Deskpro 4000 and 6000"
  *    First Edition
  *    http://h18000.www1.hp.com/support/techpubs/technical_reference_guides/113a1097.html
+ *  - IBM "Using the BIOS Build ID to identify Thinkpad systems"
+ *    Revision 2005-09-19
+ *    http://www-307.ibm.com/pc/support/site.wss/MIGR-45120.html
+ *  - Fujitsu application panel technical details
+ *    As of July 23rd, 2004
+ *    http://apanel.sourceforge.net/tech.php
  */
 
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <stdio.h>
-#include <unistd.h>
-#include <fcntl.h>
 #include <string.h>
 #include <stdlib.h>
+#include <unistd.h>
+#include <getopt.h>
 
 #include "version.h"
+#include "config.h"
 #include "types.h"
 #include "util.h"
 
-#ifdef BIGENDIAN
-typedef struct {
-	u32 h;
-	u32 l;
-} u64;
-#else /* BIGENDIAN */
-typedef struct {
-	u32 l;
-	u32 h;
-} u64;
-#endif /* BIGENDIAN */
+/* Options are global */
+struct opt
+{
+	const char* devmem;
+	unsigned int flags;
+};
+static struct opt opt;
 
-#define WORD(x) (*(const u16 *)(x))
-#define DWORD(x) (*(const u32 *)(x))
-#define QWORD(x) (*(const u64 *)(x))
+#define FLAG_VERSION            (1<<0)
+#define FLAG_HELP               (1<<1)
 
 struct bios_entry {
 	const char *anchor;
+	size_t anchor_len; /* computed */
 	off_t low_address;
 	off_t high_address;
 	size_t (*length)(const u8 *);
@@ -115,8 +116,10 @@ static int smbios_decode(const u8 *p, size_t len)
 	return 1;
 }
 
-static size_t dmi_length(__attribute__ ((unused)) const u8 *p)
+static size_t dmi_length(const u8 *p)
 {
+	(void) p;
+	
 	return(0x0F);
 }
 
@@ -241,12 +244,14 @@ static int acpi_decode(const u8 *p, size_t len)
 	
 	printf("ACPI%s present.\n",
 		acpi_revision(p[15]));
-	printf("\tOEM Identifier: ");
-	fwrite(p+9, 6, 1, stdout); 
-	printf("\n");
+	printf("\tOEM Identifier: %c%c%c%c%c%c\n",
+		p[9], p[10], p[11], p[12], p[13], p[14]);
 	printf("\tRSD Table 32-bit Address: 0x%08X\n",
 		DWORD(p+16));
 	
+	if(len<36)
+		return 1;
+	
 	if(DWORD(p+20)>len || !checksum(p, DWORD(p+20)))
 		return 0;
 	
@@ -267,9 +272,9 @@ static size_t sony_length(const u8 *p)
 	return(p[0x05]);
 }
 
-static int sony_decode(const u8 *p, __attribute__ ((unused)) size_t len)
+static int sony_decode(const u8 *p, size_t len)
 {
-	if(!checksum(p, p[0x05]))
+	if(!checksum(p, len))
 		return 0;
 	
 	printf("Sony system detected.\n");
@@ -398,21 +403,28 @@ static size_t compaq_length(const u8 *p)
 	return (p[4]*10+5);
 }
 
-static int compaq_decode(const u8 *p, __attribute__ ((unused)) size_t len)
+static int compaq_decode(const u8 *p, size_t len)
 {
-	int i;
+	unsigned int i;
+	(void) len;
 
 	printf("Compaq-specific entries present.\n");
 
 	/* integrity checking (lack of checksum) */
 	for(i=0; i<p[4]; i++)
 	{
+		/*
+		 * We do not check for truncated entries, because the length
+		 * was computed from the number of records in compaq_length
+		 * right above, so it can't be wrong.
+		 */
 		if(p[5+i*10]!='$' || !(p[6+i*10]>='A' && p[6+i*10]<='Z')
 			|| !(p[7+i*10]>='A' && p[7+i*10]<='Z')
 			|| !(p[8+i*10]>='A' && p[8+i*10]<='Z'))
 		{
-			printf("\t Abnormal Entry! Please report. [%02X %02X %02X %02X]\n",
-				p[5+i*10], p[6+i*10], p[7+i*10], p[8+i*10]);
+			printf("\t Abnormal entry! Please report. [%02X %02X "
+				"%02X %02X]\n", p[5+i*10], p[6+i*10],
+				p[7+i*10], p[8+i*10]);
 			return 0;
 		}
 	}
@@ -427,94 +439,228 @@ static int compaq_decode(const u8 *p, __attribute__ ((unused)) size_t len)
 	return 1;
 }
 
+/*
+ * VPD (vital product data, IBM-specific)
+ */
+
+static void vpd_print_entry(const char *name, const u8 *p, size_t len)
+{
+	size_t i;
+	
+	printf("\t%s: ", name);
+	for(i=0; i<len; i++)
+		if(p[i]>=32 && p[i]<127)
+			printf("%c", p[i]);
+	printf("\n");
+}
+
+static size_t vpd_length(const u8 *p)
+{
+	return (p[5]);
+}
+
+static int vpd_decode(const u8 *p, size_t len)
+{
+	if(len<0x30)
+		return 0;
+
+	/* XSeries have longer records. */
+	if(!(len>=0x45 && checksum(p, len))
+	/* Some Netvista seem to work with this. */
+	&& !checksum(p, 0x30)
+	/* The Thinkpad checksum does *not* include the first 13 bytes. */
+	&& !checksum(p+0x0D, 0x30-0x0D))
+		return 0;
+	
+	printf("VPD present.\n");
+
+	vpd_print_entry("BIOS Build ID", p+0x0D, 9);
+	vpd_print_entry("Box Serial Number", p+0x16, 7);
+	vpd_print_entry("Motherboard Serial Number", p+0x1D, 11);
+	vpd_print_entry("Machine Type/Model", p+0x28, 7);
+	
+	if(len<0x45)
+		return 1;
+
+	vpd_print_entry("BIOS Release Date", p+0x30, 8);
+	
+	return 1;
+}
+
+/*
+ * Fujitsu application panel
+ */
+
+static size_t fjkeyinf_length(const u8 *p)
+{
+	(void) p;
+	/*
+	 * We don't know at this point, it's somewhere between 12 and 32.
+	 * So we return the max, it shouldn't hurt.
+	 */
+	return 32;
+}
+
+static int fjkeyinf_decode(const u8 *p, size_t len)
+{
+	(void) len;
+	int i;
+
+	printf("Fujitsu application panel present.\n");
+
+	for (i = 0; i < 6; i++)
+	{
+		if (*(p+8+i*4)==0)
+			return 1;
+		printf("\tDevice %d: type %u, chip %u", i+1,
+		       *(p+8+i*4), *(p+8+i*4+2));
+		if (*(p+8+i*4+1)) /* Access method */
+			printf(", SMBus address 0x%x", *(p+8+i*4+3) >> 1);
+		printf("\n");
+	}
+
+	return 1;
+}
+
 /*
  * Main
  */
 
 static struct bios_entry bios_entries[]={
-	{ "_SM_", 0xF0000, 0xFFFFF, smbios_length, smbios_decode },
-	{ "_DMI_", 0xF0000, 0xFFFFF, dmi_length, dmi_decode },
-	{ "_SYSID_", 0xE0000, 0xFFFFF, sysid_length, sysid_decode },
-	{ "$PnP", 0xF0000, 0xFFFFF, pnp_length, pnp_decode },
-	{ "RSD PTR ", 0xE0000, 0xFFFFF, acpi_length, acpi_decode },
-	{ "$SNY", 0xE0000, 0xFFFFF, sony_length, sony_decode },
-	{ "_32_", 0xE0000, 0xFFFFF, bios32_length, bios32_decode },
-	{ "$PIR", 0xF0000, 0xFFFFF, pir_length, pir_decode },
-	{ "32OS", 0xE0000, 0xFFFFF, compaq_length, compaq_decode },
-	{ NULL, 0, 0, NULL, NULL }
+	{ "_SM_", 0, 0xF0000, 0xFFFFF, smbios_length, smbios_decode },
+	{ "_DMI_", 0, 0xF0000, 0xFFFFF, dmi_length, dmi_decode },
+	{ "_SYSID_", 0, 0xE0000, 0xFFFFF, sysid_length, sysid_decode },
+	{ "$PnP", 0, 0xF0000, 0xFFFFF, pnp_length, pnp_decode },
+	{ "RSD PTR ", 0, 0xE0000, 0xFFFFF, acpi_length, acpi_decode },
+	{ "$SNY", 0, 0xE0000, 0xFFFFF, sony_length, sony_decode },
+	{ "_32_", 0, 0xE0000, 0xFFFFF, bios32_length, bios32_decode },
+	{ "$PIR", 0, 0xF0000, 0xFFFFF, pir_length, pir_decode },
+	{ "32OS", 0, 0xE0000, 0xFFFFF, compaq_length, compaq_decode },
+	{ "\252\125VPD", 0, 0xF0000, 0xFFFFF, vpd_length, vpd_decode },
+	{ "FJKEYINF", 0, 0xF0000, 0xFFFFF, fjkeyinf_length, fjkeyinf_decode },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
-int main(__attribute__ ((unused)) int argc, const char *argv[])
+/* Believe it or not, this is significantly faster than memcmp and strncmp */
+static inline int anchor_match(const struct bios_entry *entry, const char *p)
 {
-	u8 buf[16];
-	int fd;
-	off_t fp=0xE0000;
-	const char *devmem="/dev/mem";
+	size_t i;
+	
+	for(i=0; i<entry->anchor_len; i++)
+		if(entry->anchor[i]!=p[i])
+			return 0;
+
+	return 1;
+}
+
+/* Return -1 on error, 0 on success */
+static int parse_command_line(int argc, char * const argv[])
+{
+	int option;
+	const char *optstring = "d:hV";
+	struct option longopts[]={
+		{ "dev-mem", required_argument, NULL, 'd' },
+		{ "help", no_argument, NULL, 'h' },
+		{ "version", no_argument, NULL, 'V' },
+		{ 0, 0, 0, 0 }
+	};
+
+	while((option=getopt_long(argc, argv, optstring, longopts, NULL))!=-1)
+		switch(option)
+		{
+			case 'd':
+				opt.devmem=optarg;
+				break;
+			case 'h':
+				opt.flags|=FLAG_HELP;
+				break;
+			case 'V':
+				opt.flags|=FLAG_VERSION;
+				break;
+			case '?':
+				return -1;
+		}
+
+	return 0;
+}
+
+static void print_help(void)
+{
+	static const char *help=
+		"Usage: biosdecode [OPTIONS]\n"
+		"Options are:\n"
+		" -d, --dev-mem FILE     Read memory from device FILE (default: " DEFAULT_MEM_DEV ")\n"
+		" -h, --help             Display this help text and exit\n"
+		" -V, --version          Display the version and exit\n";
+	
+	printf("%s", help);
+}
+
+int main(int argc, char * const argv[])
+{
+	u8 *buf;
+	off_t fp;
+	int i;
 	
 	if(sizeof(u8)!=1 || sizeof(u16)!=2 || sizeof(u32)!=4)
 	{
-		fprintf(stderr,"%s: compiler incompatibility\n", argv[0]);
+		fprintf(stderr, "%s: compiler incompatibility\n", argv[0]);
 		exit(255);
 	}
 	
-	if(argc>=2)
-		devmem=argv[1];
-	if((fd=open(devmem, O_RDONLY))==-1 || lseek(fd, fp, SEEK_SET)==-1)
+	/* Set default option values */
+	opt.devmem=DEFAULT_MEM_DEV;
+	opt.flags=0;
+
+	if(parse_command_line(argc, argv)<0)
+		exit(2);
+
+	if(opt.flags & FLAG_HELP)
 	{
-		perror(devmem);
-		exit(1);
+		print_help();
+		return 0;
+	}
+
+	if(opt.flags & FLAG_VERSION)
+	{
+		printf("%s\n", VERSION);
+		return 0;
 	}
 	
 	printf("# biosdecode %s\n", VERSION);
-	while(fp<0xFFFFF)
+
+	if((buf=mem_chunk(0xE0000, 0x20000, opt.devmem))==NULL)
+		exit(1);
+
+	/* Compute anchor lengths once and for all */
+	for(i=0; bios_entries[i].anchor!=NULL; i++)
+		bios_entries[i].anchor_len = strlen(bios_entries[i].anchor);
+
+	for(fp=0xE0000; fp<=0xFFFF0; fp+=16)
 	{
-		int i;
-		
-		if(myread(fd, buf, 16, devmem)==-1)
-			exit(1);
+		u8 *p=buf+fp-0xE0000;
 		
 		for(i=0; bios_entries[i].anchor!=NULL; i++)
 		{
-			if(strncmp((char *)buf, bios_entries[i].anchor, strlen(bios_entries[i].anchor))==0
+			if(anchor_match(&bios_entries[i], (char *)p)
 			 && fp>=bios_entries[i].low_address
 			 && fp<bios_entries[i].high_address)
 			{
-				off_t len=bios_entries[i].length(buf);
-				u8 *p;
+				off_t len=bios_entries[i].length(p);
 				
 				if(fp+len-1<=bios_entries[i].high_address)
 				{
-					if((p=malloc(len))==NULL)
-					{
-						perror("malloc");
-						exit(1);
-					}
-
-					memcpy(p, buf, 16);
-					if(len>16)
-					{
-						/* buffer completion */
-						if(myread(fd, p+16, len-16, devmem)==-1)
-						{
-							free(p);
-							exit(1);
-						}
-					}
 					if(bios_entries[i].decode(p, len))
+					{
 						fp+=(((len-1)>>4)<<4);
-					lseek(fd, fp+16, SEEK_SET);
-					free(p);
+						break;
+					}
 				}
 			}
 		}
-		fp+=16;
-	}
-	
-	if(close(fd)==-1)
-	{
-		perror(devmem);
-		exit(1);
 	}
+
+	free(buf);	
 	
 	return 0;
 }
